version = "0.0.1"
author = "Hubert Waleńczak, Nick Belterman, Amyr Lourensz, Marwa Rouah, Cristian Stîngă"

from .data import *
from .model_func import *
from .postprocessing import *
from .predict import *
from .train import *
from .utils import *



import requests
import cv2

url = "http://localhost:8000/predict"

files = {'files': open('../data/test/test/035_43-14-ROOT1-2023-08-08_pvd_OD01_f6h1_03-Fish Eye Corrected.png', 'rb')}
data = {'model_name': 'test'}

response = requests.post(url, files=files, data=data)

# The .json() method returns the JSON response as a Python dictionary
parse_respons = response.json()

print(parse_respons)



import os
import shutil
from typing import List, Optional

from fastapi import FastAPI, File, HTTPException, UploadFile
from pydantic import BaseModel

try:
    from .model_func import load_pretrained_model
    from .predict import predict
    from .train import main as train
    from .utils import setup_logger
except:
    from model_func import load_pretrained_model
    from predict import predict
    from train import main as train
    from utils import setup_logger

UPLOAD_DIRECTORY = "to_predict"

app = FastAPI()
logger = setup_logger()

# Move your existing prediction code here
# Ensure that `load_pretrained_model`, `predict`, and all other required functions are imported


class PredictionResult(BaseModel):
    predicted_mask: List[List[int]]
    mean_conf_score: float


@app.post("/predict", response_model=List[PredictionResult])
async def predict_image(
    user_folder: str = "anonymous",
    files: List[UploadFile] = File(...),
    model_name: str = "test",
) -> List[PredictionResult]:
    results = []
    try:
        
        print(files)
        for file in files:
            # Make sure folders exist
            os.makedirs(os.path.dirname(UPLOAD_DIRECTORY), exist_ok=True)
            os.makedirs(os.path.dirname(user_folder + "/images/"), exist_ok=True)
            os.makedirs(os.path.dirname(user_folder + "/masks/"), exist_ok=True)
            print('-----')
            print("Uploda directory" + UPLOAD_DIRECTORY)
            # Save the uploaded file locally
            image_file_path = os.path.join(UPLOAD_DIRECTORY, file.filename)

            with open(image_file_path, "wb") as f:
                f.write(await file.read())

            # Get filename without extension
            base = os.path.basename(file_path)
            base_file_name = os.path.splitext(base)[0]

            # Load the pretrained model
            model = load_pretrained_model(model_name)

            # Predict using your existing function and save mask in history
            predicted_mask, mean_conf_score = predict(
                model,
                file_path,
                save_path=user_folder + "/masks/" + base_file_name + "_root_mask.tif",
            )

            # Move base image to history
            shutil.move(
                image_file_path, user_folder + "/images/" + base_file_name + ".png"
            )

            # Append the prediction result
            results.append(
                PredictionResult(
                    predicted_mask=predicted_mask, mean_conf_score=mean_conf_score
                )
            )

        return results
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Prediction error: {e}")

print('dupa')


#!/bin/python3
import copy
import json
import logging
import os
import shutil
from typing import Dict, Iterable

import numpy as np
import tensorflow
import tensorflow.keras.backend as K


def setup_logger(folder: str = "log") -> None:
    """
    Set up a logger that writes log messages to a file and the console.

    This function creates a logger that writes log messages to a specified
    file and the console. The log messages include a timestamp, the logger's
    name, the severity level of the log message, and the message itself.

    Parameters:
        - folder (str): The directory where the log file will be created. Defaults to "log".

    Returns:
        logging.Logger: Configured logger instance.

    Example:
        .. code-block:: python

            logger = setup_logger()
            logger.info("This is an info message.")
    """
    # Check if logger with the same name already exists
    logger = logging.getLogger(__name__)
    if logger.handlers:
        # Logger already configured, return it
        return logger

    filename = "buas_cv6.log"
    path = os.path.join(folder, filename)
    os.makedirs(folder, exist_ok=True)

    # Create a logger object
    logger = logging.getLogger(__name__)

    # Set the logging level
    logger.setLevel(logging.INFO)

    # Create a handler for writing to a file
    file_handler = logging.FileHandler(path)

    # Create a handler for writing to the console
    console_handler = logging.StreamHandler()

    # Create a formatter and add it to the handlers
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    return logger


logger = setup_logger()


def f1(y_true: Iterable[float], y_pred: Iterable[float]) -> float:
    """
    Calculate the F1 score.

    F1 score is the harmonic mean of precision and recall.
    It's a commonly used metric in binary classification tasks.

    Parameters:
        - y_true (Iterable[float]): True labels.
        - y_pred (Iterable[float]): Predicted labels.

    Returns:
        float: The F1 score.
    """

    def recall_m(y_true: Iterable[float], y_pred: Iterable[float]) -> float:
        y_true = tensorflow.cast(y_true, tensorflow.float32)
        y_pred = tensorflow.cast(y_pred, tensorflow.float32)
        TP = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        Positives = K.sum(K.round(K.clip(y_true, 0, 1)))
        recall = TP / (Positives + K.epsilon())
        return recall

    def precision_m(y_true: Iterable[float], y_pred: Iterable[float]) -> float:
        y_true = tensorflow.cast(y_true, tensorflow.float32)
        y_pred = tensorflow.cast(y_pred, tensorflow.float32)
        TP = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        Pred_Positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
        precision = TP / (Pred_Positives + K.epsilon())
        return precision

    try:
        precision, recall = precision_m(y_true, y_pred), recall_m(y_true, y_pred)
    except ValueError:
        logger.error(
            f"An ValueError occurred while calculating precision and recall due to mismatched shapes between {y_true.shape = } and {y_pred.shape =}."
        )

    f1_score = 2 * ((precision * recall) / (precision + recall + K.epsilon()))
    return f1_score


def iou(y_true: Iterable[float], y_pred: Iterable[float]) -> float:
    """
    Calculate the Intersection over Union (IoU) score.

    Intersection over Union (IoU) is a measure used to evaluate the
    overlap between two boundaries. In the context of object detection
    or segmentation, it's used to evaluate the accuracy of predicted
    bounding boxes or segmentations against the ground truth.

    Parameters:
        - y_true (Iterable[float]): True labels.
        - y_pred (Iterable[float]): Predicted labels.

    Returns:
        float: The IoU score.
    """

    def f(y_true: Iterable[float], y_pred: Iterable[float]) -> float:
        y_true = tensorflow.cast(y_true, tensorflow.float32)
        y_pred = tensorflow.cast(y_pred, tensorflow.float32)
        intersection = K.sum(K.abs(y_true * y_pred), axis=[1, 2, 3])
        total = K.sum(K.square(y_true), [1, 2, 3]) + K.sum(K.square(y_pred), [1, 2, 3])
        union = total - intersection
        return (intersection + K.epsilon()) / (union + K.epsilon())

    try:
        iou_ = K.mean(f(y_true, y_pred), axis=-1)
        logger.debug(f"iou return value - {iou_}")
    except ValueError:
        logger.error(
            f"An ValueError occurred while calculating iou due to mismatched shapes between {y_true.shape = } and {y_pred.shape =}."
        )
    return iou_


def clear_destination_folder(image_patch_path: str) -> None:
    """
    Clears the destination folder where the image patch is located. If the folder exists,
    it is removed along with its contents. A new empty folder is then created in the same location.

    Parameters:
        - image_patch_path (str): The file path to the image patch. The folder containing this file will be cleared.

    Returns:
        None
    """
    folder_path = os.path.dirname(image_patch_path)
    if os.path.exists(folder_path):
        logger.info("Clearing destination folder")
        shutil.rmtree(folder_path)
        logger.debug(f"clear_destination_folder - Removing folder: {folder_path}")
    # Create the folder if it doesn't exist
    os.makedirs(folder_path)
    # print(f'Creating folder: {folder_path}')


def mean_confidence_score(predicted_probs: np.ndarray, threshold: float = 0.5) -> float:
    """
    Calculates the mean confidence score of the predicted probabilities that exceed a given threshold.

    Parameters:
        - predicted_probs (Union[np.ndarray, Any]): A numpy array or similar structure containing the predicted probabilities.
        - threshold (float, optional): The threshold above which the predicted probabilities are considered. Default is 0.5.

    Returns:
        - np.ndarray: The mean confidence score of the pixels above the threshold.
    """
    root_pixels = predicted_probs[predicted_probs > threshold]
    return float(np.mean(root_pixels)) if root_pixels.size > 0 else float(0)


def create_config_json(model_name: str, params: Dict) -> None:
    """
    Create or update a JSON configuration file with model parameters.

    Parameters:
        - model_name (str): The name of the model to be added or updated.
        - params (Dict): A dictionary of model's parameters.

    Returns:
        - None
    """

    # Log the action of writing parameters to the JSON file
    logger.info(
        f"main - Writing parameters to model_config.json - {model_name}: {params}"
    )

    # Define the path to the configuration JSON file
    path_config_json = r"./models/model_config.json"

    try:
        # Try load existing parameters from the JSON file if it exists
        with open(path_config_json, "r", encoding="utf-8") as json_load:
            config_dict = json.load(json_load)

        # Check if the model name already exists in the configuration
        if model_name in config_dict:
            # Log an error if the model name already exists
            logger.error(f"main - Model with name: {model_name} already exists.")
        else:
            # Add the new model parameters to the configuration dictionary
            config_dict[model_name] = params

        # Write the updated configuration back to the JSON file
        with open(path_config_json, "w", encoding="utf-8") as json_dump:
            json.dump(config_dict, json_dump, indent=4)
            logger.info(f"main - Updated config file and added model: {model_name}.")

    except FileNotFoundError:
        # If the JSON file does not exist, create a new dictionary with the model name as the key and parameters as values
        config_dict = {model_name: params}

        # Write the new configuration dictionary to a new JSON file
        with open(path_config_json, "w", encoding="utf-8") as json_file:
            json.dump(config_dict, json_file)

    except json.JSONDecodeError:
        # Handle JSON decoding error if the file is not properly formatted
        logger.error(
            f"main - JSON decode error while reading the file {path_config_json}."
        )


def read_config(model_name: str):
    """
    Reads the configuration for the specified model from a JSON file.

    Parameters:
        model_name (str): The name of the model whose configuration is to be read.

    Returns:
        dict: The configuration dictionary for the specified model.

    Raises:
        FileNotFoundError: If the configuration file does not exist.
        KeyError: If the specified model name is not found in the configuration.
        json.JSONDecodeError: If there is an error decoding the JSON file.
    """
    try:
        with open(r"models/model_config.json", "r", encoding="utf-8") as f:
            config = json.load(f)

        if model_name not in config:
            raise KeyError(f"Model '{model_name}' not found in the configuration.")

        return config[model_name][0]

    except FileNotFoundError as fnf_error:
        print(f"Error: The configuration file was not found. {fnf_error}")
        raise
    except KeyError as key_error:
        print(f"Error: {key_error}")
        raise
    except json.JSONDecodeError as json_error:
        print(f"Error: Failed to decode JSON. {json_error}")
        raise
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        raise



#!/bin/python3
import argparse
import copy
import csv
import datetime
import json
import os
from typing import Tuple

import cv2
import keras
import numpy as np
import wandb
from tensorflow.keras.callbacks import (EarlyStopping, ModelCheckpoint,
                                        TensorBoard)
from tensorflow.keras.layers import (Conv2D, Conv2DTranspose, Dropout, Input,
                                     MaxPooling2D, concatenate)
from tensorflow.keras.models import Model, load_model
from wandb.integration.keras import WandbMetricsLogger, WandbModelCheckpoint

try:
    # Attempt relative imports (if run as a package module)
    from .data import crop_to_petri, padder
    from .predict import predict
    from .utils import create_config_json, f1, iou, read_config, setup_logger

except ImportError:
    # Fallback to absolute imports (if run as a standalone script)
    from predict import predict
    from utils import create_config_json, f1, iou, read_config, setup_logger

    from data import crop_to_petri, padder

logger = setup_logger()


def create_model(
    model_name: str,
    input_shape: Tuple[int] = (256, 256, 1),
    output_classes: int = 1,
    optimizer: str = "adam",
    loss: str = "binary_crossentropy",
    output_activation: str = "sigmoid",
    dropout_1: int = 0.1,
    dropout_2: int = 0.2,
    dropout_3: int = 0.3,
    summary: bool = False,
) -> keras.models.Model:
    """
    Create a U-Net model for semantic segmentation.

    Parameters:
        - input_shape (Tuple[int]): Input shape for the model. Default is (256, 256, 1)
        - output_classes (int): Number of output classes. Default is 1.
        - optimizer (str/optimizer): Name of the optimizer to use or a custom optimizer. Default is 'adam'.
        - loss (str/loss function): Loss function to use during training. Default is 'binary_crossentropy'.
        - output_activation (str): Activation function for the output layer. Default is 'sigmoid'.
        - dropout_1 (float): Dropout rate for the first set of layers. Default is 0.1.
        - dropout_2 (float): Dropout rate for the second set of layers. Default is 0.2.
        - dropout_3 (float): Dropout rate for the third set of layers. Default is 0.3.
        - summary (bool): Whether to print the model summary. Default is False.

    Returns:
        - tensorflow.keras.models.Model: U-Net model for semantic segmentation.
    """
    # Define the logger

    # Build the model
    inputs = Input(input_shape)
    s = inputs

    # Log input shape
    logger.debug(f"Input shape: {input_shape}")

    # Contraction path
    c1 = Conv2D(
        16, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(s)
    c1 = Dropout(dropout_1)(c1)
    c1 = Conv2D(
        16, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c1)
    p1 = MaxPooling2D((2, 2))(c1)

    c2 = Conv2D(
        32, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(p1)
    c2 = Dropout(dropout_1)(c2)
    c2 = Conv2D(
        32, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c2)
    p2 = MaxPooling2D((2, 2))(c2)

    c3 = Conv2D(
        64, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(p2)
    c3 = Dropout(dropout_2)(c3)
    c3 = Conv2D(
        64, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c3)
    p3 = MaxPooling2D((2, 2))(c3)

    c4 = Conv2D(
        128, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(p3)
    c4 = Dropout(dropout_2)(c4)
    c4 = Conv2D(
        128, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c4)
    p4 = MaxPooling2D(pool_size=(2, 2))(c4)

    c5 = Conv2D(
        256, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(p4)
    c5 = Dropout(dropout_3)(c5)
    c5 = Conv2D(
        256, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c5)

    # Expansive path
    u6 = Conv2DTranspose(128, (2, 2), strides=(2, 2), padding="same")(c5)
    u6 = concatenate([u6, c4])
    c6 = Conv2D(
        128, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(u6)
    c6 = Dropout(dropout_2)(c6)
    c6 = Conv2D(
        128, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c6)

    u7 = Conv2DTranspose(64, (2, 2), strides=(2, 2), padding="same")(c6)
    u7 = concatenate([u7, c3])
    c7 = Conv2D(
        64, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(u7)
    c7 = Dropout(dropout_2)(c7)
    c7 = Conv2D(
        64, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c7)

    u8 = Conv2DTranspose(32, (2, 2), strides=(2, 2), padding="same")(c7)
    u8 = concatenate([u8, c2])
    c8 = Conv2D(
        32, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(u8)
    c8 = Dropout(dropout_1)(c8)
    c8 = Conv2D(
        32, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c8)

    u9 = Conv2DTranspose(16, (2, 2), strides=(2, 2), padding="same")(c8)
    u9 = concatenate([u9, c1], axis=3)
    c9 = Conv2D(
        16, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(u9)
    c9 = Dropout(dropout_1)(c9)
    c9 = Conv2D(
        16, (3, 3), activation="relu", kernel_initializer="he_normal", padding="same"
    )(c9)

    outputs = Conv2D(output_classes, (1, 1), activation=output_activation)(c9)

    model = Model(inputs=[inputs], outputs=[outputs])

    # Compile the model
    model.compile(optimizer=optimizer, loss=loss, metrics=["accuracy", f1, iou])

    # Show model summary
    if summary:
        logger.info(model.summary())

    # Create the model configuration dictionary
    model_config = {
        "input_shape": input_shape,
        "output_classes": output_classes,
        "optimizer": optimizer,
        "loss": loss,
        "output_activation": output_activation,
        "dropout_1": dropout_1,
        "dropout_2": dropout_2,
        "dropout_3": dropout_3,
    }
    # Add the configuration to the config file
    create_config_json(model_name, model_config)

    return model


def load_pretrained_model(model_name: str) -> keras.models.Model:
    """
    Load a saved and pre-trained U-Net model from the specified directory.

    Parameters:
        - model_name (str): The name of the model file to load.

    Returns:
        - tensorflow.keras.models.Model: The loaded U-Net model.

    Raises:
        - FileNotFoundError: If the model file does not exist at the specified path.

    Notes:
        - The model file needs to be in the './models' directory.
    """

    # Construct the model path
    model_path = os.path.join("models/", model_name + ".keras")

    # Check if the model file exists
    if not os.path.exists(model_path):
        logger.error(f"No model found at {model_path} with the name: {model_name}")
        raise FileNotFoundError(
            f"No model found at {model_path} with the name: {model_name}"
        )

    # Load the model
    model = load_model(model_path, custom_objects={"f1": f1, "iou": iou})

    # Log the model being loaded succesfully
    logger.info(f"Model loaded successfully from {model_path}")

    return model


def train_model(
    model_name: str,
    train_generator,
    test_generator,
    steps_per_epoch: int,
    validation_steps: int,
    epochs: int = 20,
    patience: int = 5,
) -> None:
    """
    Trains and saves a convolutional neural network model using the specified architecture.

    Parameters:
        - model_name (str): Name of model selected by user. This is used to retrieve the model parameters.
        - train_generator: Training data generator.
        - test_generator: Testing data generator.
        - steps_per_epoch (int): Number of steps to be taken per epoch.
        - validation_steps (int): Number of steps to be taken for validation.
        - epochs (int, optional): Number of training epochs (default is 20).
        - patience (int, optional): Number of epochs the the training loop will wait to see if the val_loss improves.

    Returns:
        None

    Notes:
        - The function initializes a neural network model based on the specified parameters.
        - Training is performed using the provided data generators and hyperparameters.
        - Early stopping and model checkpoint callbacks are applied during training.
        - The best model is saved to a file with the specified suffix.
    """
    # Check model availability, if not, create new one
    try:
        model = load_pretrained_model(model_name)
        logger.info(f"{model_name} loaded.")
    except FileNotFoundError:
        model = create_model(model_name)
        logger.info(f"{model_name} created.")

    # Format the current time
    Now = datetime.datetime.now()
    time = Now.strftime("%Y.%m.%d-%H:%M")

    # Start a run, tracking hyperparameters
    wandb.init(
        # Set the wandb project where this run will be logged
        project='npeccv6',
        # Set the name for the run
        name=model_name + "_" + time,
        # Track hyperparameters and run metadata with wandb.config
        sync_tensorboard=True,
        # Configuration
        config={
            "metric": ["accuracy", f1, iou],
            "epoch": epochs,
            "steps_per_epoch": steps_per_epoch,
        },
    )

    tb = TensorBoard(log_dir="./logs/tensorboard/" + time, histogram_freq=1)
    # Log TensorBoard directory
    logger.info(
        f'Tensorboard of {model_name} at location {"./logs/tensorboard/" + time}'
    )
    # Early stopping callback
    cb = EarlyStopping(
        monitor="val_loss", patience=patience, restore_best_weights="True", mode="min"
    )
    # Model checkpoint callback
    save_model = ModelCheckpoint(
        f"models/{model_name}.keras",
        save_best_only=True,
        monitor="val_loss",
        mode="min",
    )
    wbml = (WandbMetricsLogger(log_freq="epoch"))
    # Weights and Biases model checkpoint
    wbmc = WandbModelCheckpoint(
        filepath=f"./models/wandb/{model_name}_{time}", save_best_only=True),

    history = model.fit(
        train_generator,
        steps_per_epoch=steps_per_epoch,
        epochs=epochs,
        validation_data=test_generator,
        validation_steps=validation_steps,
        callbacks=[
            cb,
            tb,
            save_model,
            wbml,
            wbmc
        ],
    )
    # Finish the wandb run
    wandb.finish()
    # Create the models folder if it does not exist
    os.makedirs("./models/", exist_ok=True)
    # Save the trained model
    model.save("./models/" + time + ".keras")
    # Log saving the model
    logger.info(f'{model_name} saved at location {"./models/" + time}')
    # Log the model history
    logger.info(f"train_model - history - {history}")


def model_eval(
    model_name: str,
    test_data_dir: str,
) -> Tuple[float, float]:
    """
    Evaluates a pre-trained model on a test dataset of images and their corresponding masks.

    Parameters:
        - model_name (str): The name or path of the pre-trained model to be loaded.
        - test_data_dir (str): The directory containing the test images and masks.

    Returns:
        Tuple[float, float]:
            A tuple containing the F1 score and IoU (Intersection over Union) score for the model's predictions on the test dataset.

    Notes:
        The function performs the following steps:
        1. Sets up a logger for debugging and logging purposes.
        2. Initializes empty lists to store predictions (preds) and ground truth masks (y_true).
        3. Loads the specified pre-trained model.
        4. Iterates through each file in the test data directory:
            - Checks if the file is a PNG image.
            - Constructs the file path for the image and its corresponding mask.
            - Reads the mask image using OpenCV.
            - Uses the model to predict the mask for the image.
            - Appends the predicted mask and the ground truth mask to their respective lists.
        5. Calculates the F1 score and IoU score using the ground truth masks and the model's predictions.
        6. Logs and returns the F1 score and IoU score as a tuple.
    """
    # Set up logger
    logger = setup_logger()

    # Read params from config
    config = read_config(model_name)

    # Create path to masks and images
    test_data_folder_list = os.listdir(test_data_dir)
    test_data_image_path = os.path.join(test_data_dir, test_data_folder_list[0])
    test_data_mask_path = os.path.join(test_data_dir, test_data_folder_list[1])
    logger.debug(f"model_eval - test_data_image_path - {test_data_image_path}")
    logger.debug(f"model_eval - test_data_mask_path - {test_data_mask_path}")

    # Define empty list for preds
    preds = []

    # Define empty list for y true
    y_true = []

    # Define empty list for confidence score
    conf_scores = []

    # Load model
    model = load_pretrained_model(model_name)

    # Enumerate through test data directory
    for idx, file in enumerate(os.listdir(test_data_image_path)):
        # Check if file is png
        if file.endswith(".png"):
            # Create filepath to image
            file_path_image = os.path.join(test_data_dir, "images", file)
            logger.debug(f"model_eval - file_path_image - {file_path_image}")

            # Split file on . taking the first item
            file_name = file.split(".")[0]
            logger.debug(f"model_eval - file_name - {file_name}")

            # Create filepath to mask
            file_path_mask = os.path.join(
                test_data_dir, "masks", file_name + "_root_mask.tif"
            )
            logger.debug(f"model_eval - file_path_mask - {file_path_mask}")

            # Read mask as image
            mask_im = cv2.imread(file_path_mask, 0)
            logger.debug(f"model_eval - mask_im - {mask_im.shape}")

            # Read image as grayscale
            im = cv2.imread(file_path_image, 0)

            # Predict using model on a single image
            pred, conf = predict(
                model,
                file_path_image,
                r"../predictions/test_image_{}.png".format(idx),
            )

            logger.debug(f"model_eval - pred - {pred.shape}")

            # Extract petri region of interest and apply it to the mask_im
            _, roiX, roiY, roiW, roiH = crop_to_petri(im)
            cropped_mask = mask_im[roiY : roiY + roiH, roiX : roiX + roiW]
            logger.debug(f"model_eval - cropped_mask.shape - {cropped_mask.shape}")

            # Adding padding to the mask
            padded_cropped_mask = padder(cropped_mask, config["input_shape"][0])

            # Append items to lists
            preds.append(pred)
            y_true.append(padded_cropped_mask)
            conf_scores.append(conf)

    # Convert lists to numpy arrays
    preds = np.array(preds)
    y_true = np.array(y_true)
    conf_scores = np.array(conf_scores)
    logger.debug(f"model_eval - preds.shape - {preds.shape}")
    logger.debug(f"model_eval - y_true.shape - {y_true.shape}")
    logger.debug(f"model_eval - conf_scores.shape - {conf_scores.shape}")

    # Calculate f1
    f1_ = f1(y_true, preds)
    logger.info(f"model_eval - f1 - {f1_}")

    # Calculate iou
    iou_ = iou(y_true, preds)
    logger.info(f"model_eval - iou - {iou_}")

    avg_conf_over_all_preds = np.mean(conf_scores)

    return f1_, iou_, avg_conf_over_all_preds


def main() -> None:
    parser = argparse.ArgumentParser(description="Create model_config.json.")

    parser.add_argument(
        "--input_shape",
        nargs="+",
        type=int,
        default=[256, 256, 1],
        help="Input shape for the model. Default is (256, 256, 1)",
    )
    parser.add_argument(
        "--output_classes",
        type=int,
        default=1,
        help="Number of output classes. Default is 1.",
    )
    parser.add_argument(
        "--optimizer",
        type=str,
        default="adam",
        help="Name of the optimizer to use or a custom optimizer. Default is 'adam'.",
    )
    parser.add_argument(
        "--loss",
        type=str,
        default="binary_crossentropy",
        help="Loss function to use during training. Default is 'binary_crossentropy'.",
    )
    parser.add_argument(
        "--output_activation",
        type=str,
        default="sigmoid",
        help="Activation function for the output layer. Default is 'sigmoid'.",
    )
    parser.add_argument(
        "--dropout_1",
        type=float,
        default=0.1,
        help="Dropout rate for the first set of layers. Default is 0.1.",
    )
    parser.add_argument(
        "--dropout_2",
        type=float,
        default=0.2,
        help="Dropout rate for the second set of layers. Default is 0.2.",
    )
    parser.add_argument(
        "--dropout_3",
        type=float,
        default=0.3,
        help="Dropout rate for the third set of layers. Default is 0.3.",
    )
    parser.add_argument(
        "--summary",
        action="store_true",
        help="Whether to print the model summary. Default is False.",
    )
    parser.add_argument(
        "--model_name",
        type=str,
        help="Name of model selected by user. This is used to retrieve & store the model parameters.",
    )

    args = parser.parse_args()

    # Store parameters in JSON file
    params = vars(args)

    create_config_json(params=params)


if __name__ == "__main__":
    main()



#!/bin/python3
import argparse
from typing import Tuple

import cv2
import keras
import numpy as np
from patchify import patchify, unpatchify
from tensorflow.keras.models import load_model
from tensorflow.keras.utils import custom_object_scope

try:
    # Attempt relative imports (if run as a package module)
    from .data import crop_to_petri, padder
    from .postprocessing import process_image_for_roots
    from .utils import f1, iou, mean_confidence_score, setup_logger

except ImportError:
    # Fallback to absolute imports (if run as a standalone script)
    from postprocessing import process_image_for_roots
    from utils import f1, iou, mean_confidence_score, setup_logger

    from data import crop_to_petri, padder

logger = setup_logger()


def load_and_preprocess_image(
    im_path: str,
    patch_size: int = 256,
    scaling_factor: float = 1,
    num_channels: int = 1,
) -> Tuple[np.ndarray, int, int, np.ndarray]:
    """
    Load and preprocess an image for patch-based prediction.

    This function reads an image from the specified path, converts it to grayscale
    if required, crops it to the region of interest (ROI) determined by the `crop_to_petri`
    function, pads the image to the specified patch size, scales it, and finally divides
    the image into non-overlapping patches of the specified size.

    Parameters:
        - im_path (str): Path to the input image.
        - patch_size (int, optional): Size of the patches to extract from the image. Default is 256.
        - scaling_factor (float, optional): Factor by which to scale the image. Default is 1.
        - num_channels (int, optional): Number of channels to read from the image (1 for grayscale, 3 for color). Default is 1.

    Returns:
        tuple: A tuple containing:
            - patches (numpy.ndarray): Array of image patches with shape (num_patches, patch_size, patch_size, 1).
            - i (int): Number of patches along the height of the image.
            - j (int): Number of patches along the width of the image.
            - im (numpy.ndarray): The preprocessed image after cropping, padding, and scaling.

    Example:
        .. code-block:: python

            patches, i, j, im = load_and_preprocess_image('path/to/image.jpg', patch_size=128, scaling_factor=1, num_channels=3)
            print(patches.shape)
            (num_patches, 128, 128, 1)
            print(i, j)
            4 4
    """

    logger.info(f"Loading image from path: {im_path}")

    if num_channels == 1:
        im = cv2.imread(im_path, cv2.IMREAD_GRAYSCALE)
        logger.debug(f"load_and_preprocess_image:number of channels:{num_channels}")
        logger.debug(f"load_and_preprocess_image:im.shape:{im.shape}")
    else:
        im = cv2.imread(im_path, num_channels)
        logger.debug(f"load_and_preprocess_image:number of channels:{num_channels}")
        logger.debug(f"load_and_preprocess_image:im.shape:{im.shape}")

    if im is None:
        logger.error("Error loading image.")
        raise ValueError("Error loading image")

    if num_channels != 1:
        im_gray = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
    else:
        im_gray = im

    logger.debug(f"load_and_preprocess_image:im_gray.shape:{im_gray.shape}")

    # Crop image
    _, roiX, roiY, roiW, roiH = crop_to_petri(im_gray)
    logger.debug(
        f"load_and_preprocess_image:output crop_petri:{roiX = }\n{roiY = }\n{roiW = }\n{roiH = }"
    )
    logger.debug(
        f"load_and_preprocess_image:image slice:[{roiY}:{roiY + roiH}, {roiX}:{roiX + roiW}]"
    )

    # Pad image
    im = im[roiY : roiY + roiH, roiX : roiX + roiW]

    padded_im = padder(im, patch_size)

    logger.debug(
        f"load_and_preprocess_image:im.shape:after ROI extraction & padder:{im.shape}"
    )

    if scaling_factor != 1:
        logger.info(f"Scaling image by a factor of {scaling_factor}.")
        padded_im = cv2.resize(padded_im, (0, 0), fx=scaling_factor, fy=scaling_factor)

    logger.debug(f"{len((patch_size, patch_size) * im.ndim)} - {im.ndim}")
    # if num_channels == 1 and padded_im.ndim == 2:
    #     padded_im = np.expand_dims(padded_im, axis=-1)

    # Create patches
    try:
        patches = patchify(padded_im, (patch_size, patch_size), step=patch_size)
    except ValueError as e:
        logger.error(f"Error in patchify: {e}")
        logger.debug(
            f"Image shape for patchify: {padded_im.shape}, Patch shape: {(patch_size, patch_size, num_channels)}"
        )
        raise

    logger.debug(f"load_and_preprocess_iamge:patches.shape:{patches.shape}")

    i = patches.shape[0]
    j = patches.shape[1]

    patches = patches.reshape(-1, patch_size, patch_size, 1)
    logger.debug(f"load_and_preprocess_image:reshaped patches:{patches.shape}")

    return patches, i, j, padded_im


def postprocess_prediction(
    preds: np.ndarray,
    i: int,
    j: int,
    im: np.ndarray,
    threshold: float = 0.8,
    patch_size: int = 256,
) -> np.ndarray[int]:
    """
    Post-process the model's predictions to create a final mask.

    Parameters:
        - preds (np.ndarray): Predictions from the model.
        - i (int): Number of patches in the vertical dimension.
        - j (int): Number of patches in the horizontal dimension.
        - im (np.ndarray): Original image.
        - threshold (float): Threshold value for binarizing the mask. Default is 0.8.
        - patch_size (int): Size of the patches. Default is 256.

    Returns:
        - np.ndarray: Post-processed binary mask.
    """

    logger.info("Starting post-processing of predictions.")

    # Reshape predictions
    preds = preds.reshape(i, j, patch_size, patch_size)

    # Splice predictions
    predicted_mask = unpatchify(preds, (im.shape[0], im.shape[1]))
    predicted_mask = np.where(predicted_mask > threshold, 1, 0)

    logger.info("Post-processing completed.")
    return predicted_mask


def predict(
    model: keras.models.Model,
    im_path: str,
    save_path: str = "Predictions/Image.png",
    patch_size: int = 256,
    threshold: float = 0.8,
    scaling_factor: float = 1,
    num_channels: int = 1,
) -> Tuple[np.ndarray[int], float]:
    """
    Predict and post-process the mask for the given image.

    Parameters:
        - model (Model): Trained Keras model for prediction.
        - im_path (str): Path to the input image.
        - save_path (str): Path to save the predicted mask. Default is "Predictions/Image.png".
        - patch_size (int): Size of the patches. Default is 256.
        - threshold (float): Threshold value for binarizing the mask. Default is 0.8.
        - scaling_factor (float): Scaling factor for the image. Default is 1.
        - num_channels (int): Number of channels for the image (1 for grayscale, 3 for color). Default is 1.

    Returns:
        - np.ndarray: Predicted binary mask.
        - float: Mean confidence score of prediction
    """

    logger.info(f"Loading and preprocessing image from path: {im_path}")

    # Preprocess image
    patches, i, j, im = load_and_preprocess_image(
        im_path, patch_size, scaling_factor, num_channels
    )

    logger.info("Image preprocessing completed. Starting prediction.")

    # Predict
    preds = model.predict(patches / 255)

    # Calculate mean confidence score
    mean_conf_score = mean_confidence_score(preds, threshold)

    logger.info("Prediction completed. Starting post-processing.")

    # Postprocess prediction
    predicted_mask = postprocess_prediction(preds, i, j, im, threshold, patch_size)

    logger.info(f"Saving predicted mask to: {save_path}")

    # Save predicted mask
    cv2.imwrite(
        save_path, predicted_mask.astype(np.uint8) * 255
    )  # Convert binary mask to uint8 image

    logger.info("Predicted mask saved successfully.")
    return predicted_mask, mean_conf_score


# Example usage
# Define a dummy model for demonstration purposes (replace with actual model)
# class DummyModel:
#     def predict(self, patches):
#         # Dummy prediction logic
#         return np.random.rand(*patches.shape)

# model = DummyModel()
# image_path = 'path_to_your_image.png'
# save_path = 'Predictions/Image.png'

# predicted_mask = predict(model, image_path, save_path)


def main():
    parser = argparse.ArgumentParser(
        description="Predict a root mask from an image using a trained U-net model."
    )
    parser.add_argument("image_path", help="Path to the input image file.")
    parser.add_argument("save_path", help="Where to save predicted mask.")
    parser.add_argument(
        "-m",
        "--model",
        default="main",
        action="store",
        help="What model name to use. Default is main",
    )
    parser.add_argument(
        "-p",
        "--patch_size",
        # dest = "patch_size",
        default=256,
        action="store",
        help="How to patch the images for prediction. Pre-trained models use 256.",
    )
    parser.add_argument(
        "-t",
        "--threshold",
        # dest = "threshold",
        default=0.8,
        action="store",
        help="Threshold for the predicted mask. Defult is 0.8.",
    )

    parser.add_argument(
        "-s",
        "--scaling_factor",
        # dest = "scaling_factor",
        default=1,
        action="store",
        help="Scaling factor for the image. Default is 1",
    )
    parser.add_argument(
        "-n",
        "--num_channels",
        # dest = "num_channels",
        default=1,
        action="store",
        help="Number of channels to use with image. Default is 1",
    )

    args = parser.parse_args()

    # Load the saved model
    with custom_object_scope({"f1": f1, "iou": iou}):
        model = load_model("./models/" + args.model + ".keras")

    # Predict the mask
    predicted_mask, mean_conf_score = predict(
        model,
        args.image_path,
        args.save_path,
        args.patch_size,
        args.threshold,
        args.scaling_factor,
        args.num_channels,
    )
    # TO-DO: postprocessing
    # TO-DO: extract information about number of expected plants
    # PLACEHOLDER!!!
    expected_nr_plants = 5

    root_lengths, root_tip_coords, marked_image = process_image_for_roots(
        predicted_mask, expected_nr_plants
    )
    logger.info(f"Mean confidence score: {mean_conf_score}")
    logger.info(f"root lengths: {root_lengths}")
    logger.info(f"root tips coordinates in image (px): {root_tip_coords}")
    cv2.imwrite(args.save_path, marked_image)


if __name__ == "__main__":
    main()



#!/bin/python3
import math
from typing import Any, Dict, List, Tuple

import cv2
import networkx as nx
import numpy as np
import pandas as pd
from skan import Skeleton, summarize
from skan.csr import skeleton_to_csgraph
from skimage.morphology import (remove_small_holes, remove_small_objects,
                                skeletonize)

try:
    # Attempt relative imports (if run as a package module)
    from .utils import setup_logger

except ImportError:
    # Fallback to absolute imports (if run as a standalone script)
    from utils import setup_logger


# Configure logger
logger = setup_logger()


def segment_roots(image: np.ndarray, expected_nr_plants: int) -> np.ndarray:
    """
    The function segments roots from the input image using connected component analysis and size-based filtering.
    It first identifies connected components in the image and removes small objects and small holes.
    Then, it filters the remaining components based on their position and size.
    If the number of segmented roots exceeds the expected number of plants, only the largest ones are retained.
    Finally, the function returns the segmented roots with labeled regions.

    Parameters:
        - image (np.ndarray): An array representing the input image containing plant roots.
        - expected_nr_plants (int): The expected number of plants to be found.

    Returns:
        - np.ndarray: An array representing the segmented roots.

    Example:
        - segmented_roots_image = segment_roots(image = image_to_be_segmented, expected_nr_plants = 5)

    Note:
        - This function assumes that the input image contains plant roots.
        - This function will consider overlaping roots as one singular root.
        - The function will ignore roots smaller than 64 pixels.
    """
    if image.dtype != np.uint8:
        image = image.astype(np.uint8)
    # Get components
    _, labels, stats, _ = cv2.connectedComponentsWithStats(image)

    # Remove objects with area smaller than 64
    removed_small_obj = remove_small_objects(labels, 64)
    removed_small_obj[removed_small_obj != 0] = 1
    removed_small_obj = removed_small_obj.astype(np.uint8)
    _, labels, stats, _ = cv2.connectedComponentsWithStats(removed_small_obj)

    # Make labeles binary
    labels = labels > 0

    # Remove holes with area smalle than 64
    removed_small_holes = remove_small_holes(labels, 64)
    removed_small_holes[removed_small_holes != 0] = 1
    removed_small_holes = removed_small_holes.astype(np.uint8)
    _, labels, stats, centroids = cv2.connectedComponentsWithStats(removed_small_holes)
    # Round the centroids
    centroids = np.round(centroids).astype(int)
    # Combine stats and centroids
    combined_stats = np.hstack((stats, centroids))

    # Add id to clean canvas to only roots
    row_indices = np.arange(len(combined_stats))
    stats_with_id = np.hstack((combined_stats, row_indices.reshape(-1, 1)))

    # Empty list with roots
    roots = []

    # Sort the array
    sorted_stats = stats_with_id[stats_with_id[:, 3].argsort()[::-1]]

    # Define conditions to filter the image
    above_150_condition = sorted_stats[:, 1] > 150
    below_half_y_condition = sorted_stats[:, 1] < (labels.shape[1] / 2)
    combined_condition = above_150_condition & below_half_y_condition

    # Apply the combined condition to filter rows
    filtered_stats = sorted_stats[combined_condition]
    sorted_roots = filtered_stats[filtered_stats[:, 3].argsort()[::-1]]

    # Remove elements whose centroids are below 450
    centroids_below_450 = sorted_roots[:, 6] > 450
    centroids_above_1800 = sorted_roots[:, 6] < 1800
    centroids_condition = centroids_below_450 & centroids_above_1800

    # Keep the roots which meet the centroid condition
    sorted_roots = sorted_roots[centroids_condition]

    # Sort the roots by size in descending order
    sorted_roots = sorted(sorted_roots, key=lambda x: x[4])
    sorted_roots = np.array(sorted_roots)

    # Empty list to keep the good roots
    filtered_roots = []

    # Iterate over all roots
    for i, test_root in enumerate(sorted_roots):
        # Set up the 2 checks
        keep_1 = True
        keep_2 = True
        # Go over all the root bigger than the test root
        for j, root in enumerate(sorted_roots[i + 1 :], start=i + 1):
            # If the roots overlap vertically, failed check
            if root[0] <= test_root[5] <= root[0] + root[2]:
                keep_1 = False
            # If the roots centroids are too close, failed check
            if (
                math.sqrt((root[5] - test_root[5]) ** 2 + (root[6] - test_root[6]) ** 2)
                < 300
            ):
                keep_2 = False
        # If both checks passed, save the root
        if keep_1 and keep_2:
            filtered_roots.append(test_root)

    # Sort the roots by size in descending order and make them into an array
    roots = sorted(filtered_roots, key=lambda x: x[4], reverse=True)
    roots = np.array(roots)

    # Leave only the expected nr of plants based on size and overwrite rest
    if len(roots) > expected_nr_plants:
        roots = roots[:expected_nr_plants, :]

    # Make the mask based on the roots
    mask = np.isin(labels, roots[:, -1])

    # Create the labels/big picture
    labels[~mask] = 0
    roots = roots[roots[:, 0].argsort()]

    try:
        # Get the width and size of the 2nd biggest component
        _, _, width_1, _, size_1, _, _, _ = roots[0]
        # Get the size of the 2nd biggest component
        size_2 = roots[1][5]

        # Check the width
        if width_1 > labels.shape[0] / (expected_nr_plants - 1) and size_1 / 2 > size_2:
            logger.error("Image has overlapping plants!")
            return labels
    except Exception as e:
        logger.error(f"An error occurred: {e}")

    # Check the number of roots
    if len(roots) != expected_nr_plants:
        logger.error(f"Number of detected plants is not {expected_nr_plants}!")
        logger.error(f"Detected: {len(roots)} plants")
        return labels

    # Return the image with the labels and segmented plants
    return labels


def calculate_root_length(
    subgraph: nx.Graph, subgraph_edges: pd.DataFrame
) -> Tuple[float, float, int]:
    """
    Calculate the main and lateral root lengths and the destination node of the main root in a subgraph.

    Parameters:
        - subgraph (nx.Graph): The subgraph representing a skeleton.
        - subgraph_edges (pd.DataFrame): The edge data for the subgraph.

    Returns:
        - Tuple[float, float, int]: The main root length, lateral root length, and the destination node of the main root.
    """
    logger.info("Calculating root lengths for a subgraph.")

    # Find the node that corresponds to the lowest value of image-coord-dst-0 and image-coord-src-0
    highest_dst_node = subgraph_edges.loc[
        subgraph_edges["image-coord-dst-0"].idxmin(), "node-id-dst"
    ]
    highest_src_node = subgraph_edges.loc[
        subgraph_edges["image-coord-src-0"].idxmin(), "node-id-src"
    ]

    highest_node = min(highest_dst_node, highest_src_node)
    logger.debug(f"Starting node for root length calculation: {highest_node}")

    # Compute the lengths using single-source shortest path from the chosen node
    lengths = nx.single_source_dijkstra_path_length(
        subgraph, highest_node, weight="branch-distance"
    )

    main_root_length = max(lengths.values())
    main_root_src = highest_node
    main_root_dest = max(lengths, key=lengths.get)

    lateral_root_length = sum(lengths.values()) - main_root_length

    logger.debug(
        f"Main root length: {main_root_length}, Lateral root length: {lateral_root_length}, Main root destination: {main_root_dest}"
    )

    return main_root_length, lateral_root_length, main_root_src, main_root_dest


def calculate_root_lengths(graph: pd.DataFrame) -> Dict[int, Dict[str, float]]:
    """
    Calculate the root lengths for all unique skeleton IDs in the graph.

    Parameters:
        - graph (pd.DataFrame): The edge data for the entire graph.

    Returns:
        - Dict[int, Dict[str, float]]: A dictionary mapping skeleton IDs to their respective root lengths and main root destinations.
    """
    logger.info("Calculating root lengths for all skeletons.")

    unique_values = graph["skeleton-id"].unique()
    results = {}

    for skeleton_id in unique_values:
        logger.info(f"Processing skeleton ID: {skeleton_id}")

        subgraph_edges = graph[graph["skeleton-id"] == skeleton_id]
        subgraph = nx.from_pandas_edgelist(
            subgraph_edges,
            source="node-id-src",
            target="node-id-dst",
            edge_attr="branch-distance",
        )

        (
            main_root_length,
            lateral_root_length,
            main_root_src,
            main_root_dest,
        ) = calculate_root_length(subgraph, subgraph_edges)

        results[skeleton_id] = {
            "main_root_length": main_root_length,
            "lateral_root_length": lateral_root_length,
            "main_root_src": main_root_src,
            "main_root_dest": main_root_dest,
        }

    return results


def mark_landmarks(
    image: Any,
    nodes_summary: pd.DataFrame,
    main_root_src: List[int],
    main_root_dest: List[int],
) -> Any:
    """
    Marks landmarks on the image based on node coordinates extracted from the skeleton summary DataFrame.

    Parameters:
        - image (Any): The image on which to draw the landmarks.
        - nodes_summary (pd.DataFrame): DataFrame containing summary of skeleton nodes.
        - main_root_src (List[int]): List of node IDs for main root sources.
        - main_root_dest (List[int]): List of node IDs for main root destinations.

    Returns:
        Any: The image with landmarks drawn on it.
    """
    logger.info("Marking landmarks on the image.")

    image = np.clip(image * 300, 0, 255).astype(np.uint8)
    image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

    # Define colors
    blue = (255, 0, 0)
    red = (0, 0, 255)
    green = (0, 255, 0)

    # Extract, lateral roots, and junctions from the summary
    lateral_roots = nodes_summary.loc[
        nodes_summary["branch-type"] == 1, "node-id-dst"
    ].tolist()
    junctions = nodes_summary.loc[
        nodes_summary["branch-type"] == 2, "node-id-src"
    ].tolist()

    # Remove main_root_src and main_root_dest from list to mark as junction or lateral tip
    for src, dest in zip(main_root_src, main_root_dest):
        if src in junctions:
            junctions.remove(src)
        if dest in junctions:
            junctions.remove(dest)
        if src in lateral_roots:
            lateral_roots.remove(src)
        if dest in lateral_roots:
            lateral_roots.remove(dest)

    # Draw main roots
    for src, dest in zip(main_root_src, main_root_dest):
        for node in [src, dest]:
            # Define if the start/end is branch or endpoint and act accordingly
            if node in nodes_summary["node-id-src"].values:
                coord = nodes_summary.loc[
                    nodes_summary["node-id-src"] == node, ["coord-src-0", "coord-src-1"]
                ].values[0]
            elif node in nodes_summary["node-id-dst"].values:
                coord = nodes_summary.loc[
                    nodes_summary["node-id-dst"] == node, ["coord-dst-0", "coord-dst-1"]
                ].values[0]
            image = cv2.circle(image, (int(coord[1]), int(coord[0])), 15, blue, 5)

    # Draw lateral roots
    for node in lateral_roots:
        coord = nodes_summary.loc[
            nodes_summary["node-id-dst"] == node, ["coord-dst-0", "coord-dst-1"]
        ].values[0]
        image = cv2.circle(image, (int(coord[1]), int(coord[0])), 15, red, 5)

    # Draw junctions
    for node in junctions:
        coord = nodes_summary.loc[
            nodes_summary["node-id-src"] == node, ["coord-src-0", "coord-src-1"]
        ].values[0]
        image = cv2.circle(image, (int(coord[1]), int(coord[0])), 15, green, 5)

    return image


def process_image_for_roots(
    image: np.ndarray, expected_nr_plants: int = 5
) -> Tuple[pd.DataFrame, np.ndarray]:
    """
    Processes an image to segment roots, calculate root lengths, and mark landmarks.

    Parameters:
        - image (np.ndarray): Mask of the roots.
        - expected_nr_plants (int): expected number of plants withing the petri dish

    Returns:
        - Tuple[pd.DataFrame, np.ndarray]: A DataFrame with root lengths and the image with landmarks.
    """

    # Segment roots from the image
    segmented_roots = segment_roots(
        image, expected_nr_plants
    )  # Example value for expected_nr_plants

    # Create root skeletons and supparize them
    root_skeleton = skeletonize(segmented_roots)
    summary = summarize(Skeleton(root_skeleton))

    # Calculate root lengths
    root_lengths = calculate_root_lengths(summary)

    # Create a summary DataFrame for nodes
    nodes_summary = pd.DataFrame(summary)

    # Extract main root source and destination IDs
    main_root_src = [info["main_root_src"] for info in root_lengths.values()]
    main_root_dest = [info["main_root_dest"] for info in root_lengths.values()]

    # Mark landmarks on the original image
    marked_image = mark_landmarks(image, nodes_summary, main_root_src, main_root_dest)

    root_tip_coords = []

    for node_id in main_root_dest:
        coords = nodes_summary[nodes_summary["node-id-dst"] == node_id][
            ["coord-dst-0", "coord-dst-1"]
        ].values.tolist()
        root_tip_coords.extend(coords)

    return root_lengths, root_tip_coords, marked_image



#!/bin/python3
import argparse

try:
    # Attempt relative imports (if run as a package module)
    from .data import load_and_preprocess_data
    from .model_func import train_model
    from .utils import read_config, setup_logger

except ImportError:
    # Fallback to absolute imports (if run as a standalone script)
    from model_func import train_model
    from utils import read_config, setup_logger

    from data import load_and_preprocess_data

logger = setup_logger()


def main():
    parser = argparse.ArgumentParser(
        description="Train a U-net model from images and masks."
    )
    parser.add_argument("model_name", help="name of the model.")
    parser.add_argument(
        "-c",
        "--classes",
        default=["root"],
        help="Classes to use to train the model the model. For single class(recomended): ['class'], for multiclass(not advised):['class1', 'class2']",
    )
    parser.add_argument(
        "-d",
        "--patch_dir",
        default="./data_patched/",
        help="Path to data root directory, should end with '/'. Default: './data_patched/'",
        type=str,
    )

    parser.add_argument(
        "-s",
        "--seed",
        default=42,
        help="Seed for reading data and model training",
        type=int,
    )
    parser.add_argument(
        "-b",
        "--batch_size",
        default=16,
        help="What batch size to use. Default: 16",
        type=int,
    )

    parser.add_argument(
        "-e", "--epochs", default=20, help="Number of epochs. Default: 20", type=int
    )
    # Disabled
    parser.add_argument(
        "-r",
        "--roots",
        default=5,
        help="DISABLED!!! Number of expected plants inside petri dish. Default: 5",
    )

    args = parser.parse_args()
    model_name = args.model_name
    config = read_config(model_name)

    (
        train_generator,
        test_generator,
        steps_per_epoch,
        validation_steps,
    ) = load_and_preprocess_data(
        classes=args.classes,
        model_name=model_name,
        patch_size=config["input_shape"][0],
        patch_dir=args.patch_dir,
        seed=args.seed,
        batch_size=args.batch_size,
    )

    logger.info(f"Training model\n{steps_per_epoch = }\n{validation_steps = }")
    logger.info(config["input_shape"])

    train_model(
        model_name=model_name,
        train_generator=train_generator,
        test_generator=test_generator,
        steps_per_epoch=steps_per_epoch,
        validation_steps=validation_steps,
        epochs=args.epochs,
    )


if __name__ == "__main__":
    main()



from fastapi import FastAPI
from model_func import create_model

app = FastAPI()


@app.get("/")
def read_root():
    return {"Hello": "World"}


@app.get("/create_model/{name}")
def create_model_api(name: str):
    create_model(name)



#!/bin/python3
import argparse
import glob
import os
import random
from typing import List, Tuple

import cv2
import numpy as np
from patchify import patchify
# from tensorflow.keras.models import Model
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.utils import Sequence

try:
    # Attempt relative imports (if run as a package module)
    from .utils import clear_destination_folder, setup_logger
except ImportError:
    # Fallback to absolute imports (if run as a standalone script)
    from utils import clear_destination_folder, setup_logger

logger = setup_logger()


def compare_and_remove(folder1: str, folder2: str) -> None:
    """
    Compares files in two folders and removes files from the first folder
    that don't have corresponding matches in the second folder.

    Parameters:
        - folder1 (str): Path to the first folder containing files with extension .png.
        - folder2 (str): Path to the second folder containing files with various extensions such as '_root_mask.tif', '_occluded_root_mask.tif', '_shoot_mask.tif', '_seed_mask.tif'.

    Notes:
        - The function extracts filenames without extensions from folder1 and folder2.
        - It identifies files in folder1 that do not have corresponding matches in folder2.
        - Files without matches in folder2 are removed from folder1.
        - The function prints a message for each removed file.
    """

    # Get filenames in folder1 with extension .png
    png_files = [
        file.replace(".png", "")
        for file in os.listdir(folder1)
        if file.endswith(".png")
    ]
    logger.info(f"png_files lenght: {len(png_files)}")

    # Get filenames in folder2 with various extensions
    tif_files = [
        file.replace("_root_mask.tif", "")
        .replace("_occluded_root_mask.tif", "")
        .replace("_shoot_mask.tif", "")
        .replace("_seed_mask.tif", "")
        for file in os.listdir(folder2)
        if file.endswith(
            (
                "_root_mask.tif",
                "_occluded_root_mask.tif",
                "_shoot_mask.tif",
                "_seed_mask.tif",
            )
        )
    ]

    logger.info(f"tif_files lenght: {len(tif_files)}")

    # Find files that don't match
    files_to_remove = set(png_files) - set(tif_files)

    logger.info(
        f"Lenght files_to_remove: {len(files_to_remove)}\nFiles to remove: {files_to_remove}"
    )

    # Remove files from folder1
    for file_to_remove in files_to_remove:
        file_path = os.path.join(folder1, file_to_remove + ".png")
        os.remove(file_path)

        logger.info(f"Removed: {file_path}")


def crop_to_petri(im: np.ndarray) -> Tuple[np.ndarray, int, int, int, int]:
    """
    Detect and crop the petri dish from a grayscale image.

    Parameters:
        - im (np.ndarray): Input image (grayscale).

    Returns:
        - crop_im (np.ndarray): Cropped image containing the petri dish.
        - x (int): X-coordinate of the top-left corner of the petri dish ROI.
        - y (int): Y-coordinate of the top-left corner of the petri dish ROI.
        - w (int): Width of the petri dish ROI.
        - h (int): Height of the petri dish ROI.
    """
    logger.debug("Starting petri dish detection and cropping.")
    _, output_im = cv2.threshold(im, 100, 255, cv2.THRESH_BINARY)
    _, labels, stats, centroids = cv2.connectedComponentsWithStats(output_im)
    stats = stats[1:]
    max_area_index = np.argmax(stats[:, 4])
    x, y, w, h, _ = stats[max_area_index]
    side_length = min(w, h)
    x2 = x + side_length
    y2 = y + side_length
    crop_im = im[y:y2, x:x2]
    logger.debug("Petri dish cropping completed.")
    return crop_im, x, y, side_length, side_length
    # # Correct usage
    # image_path = 'youre_image_path_here'
    # image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)  # Ensure the image is read in grayscale

    # # Check if image was loaded successfully
    # if image is None:
    #     print("Error loading image")
    # else:
    #     cropped_image, x, y, w, h = crop_to_petri(image)
    #     print(f"Cropped region coordinates: x={x}, y={y}, width={w}, height={h}")
    #     # Optionally display the cropped image
    #     # cv2.imshow("Cropped Petri Dish", cropped_image)
    #     # cv2.waitKey(0)
    #     # cv2.destroyAllWindows()

    # cropped_image.shape


def padder(image: np.ndarray, patch_size: int) -> np.ndarray:
    """
    Adds padding to an image to make its dimensions divisible by a specified patch size.

    This function calculates the amount of padding needed for both the height and width of an image
    so that its dimensions become divisible by the given patch size. The padding is applied evenly to both sides
    of each dimension (top and bottom for height, left and right for width). If the padding amount is odd,
    one extra pixel is added to the bottom or right side. The padding color is set to black (0, 0, 0).

    Parameters:
        - image (np.ndarray): The input image as a NumPy array. Expected shape is (height, width, channels).
        - patch_size (int): The patch size to which the image dimensions should be divisible. It's applied to both height and width.

    Returns:
        - np.ndarray: The padded image as a NumPy array with the same number of channels as the input.
          Its dimensions are adjusted to be divisible by the specified patch size.

    Example:
        - padded_image = padder(cv2.imread('example.jpg'), 128)
    """
    logger = setup_logger()
    logger.debug(
        f"Starting padding for image of shape {image.shape} to make dimensions divisible by {patch_size}."
    )

    h, w = image.shape[:2]
    height_padding = ((h // patch_size) + 1) * patch_size - h
    width_padding = ((w // patch_size) + 1) * patch_size - w

    top_padding = height_padding // 2
    bottom_padding = height_padding - top_padding

    left_padding = width_padding // 2
    right_padding = width_padding - left_padding

    logger.debug(
        f"Calculated padding: top={top_padding}, bottom={bottom_padding}, left={left_padding}, right={right_padding}"
    )

    padded_image = cv2.copyMakeBorder(
        image,
        top_padding,
        bottom_padding,
        left_padding,
        right_padding,
        cv2.BORDER_CONSTANT,
        value=[0, 0, 0],
    )

    logger.debug(f"Padding completed. New image shape is {padded_image.shape}.")

    return padded_image


def save_mask_patches(
    path: str,
    mask: np.ndarray,
    patch_size: int,
    root_path: str,
    subdir: str,
    patch_dir: str,
    dataset_type: str,
    i: int,
) -> None:
    """
    Supplemental function for patch_and_save.
    Creates and saves patches for a given mask image.

    Parameters:
        - path (str): Path to the original mask image file.
        - mask (numpy.ndarray): Mask image to create patches from.
        - patch_size (int): Size of the patches to be created.
        - subdir (str): Subdirectory within the dataset.
        - patch_dir (str): Directory where patches will be saved.
        - dataset_type (str): Type of the dataset (e.g., 'train', 'test').
        - i (int): Index used for saving numbered patches.

    Notes:
        - Utilizes the patchify method to create patches from the mask image.
        - Reshapes and saves patches to the specified patch directory with numbered filenames.
    """
    # Create patches for mask
    patches = patchify(mask, (patch_size, patch_size, 1), step=patch_size)
    patches = patches.reshape(-1, patch_size, patch_size, 1)

    # Make a correct path to save
    image_patch_path = path.replace(root_path, patch_dir)
    # image_patch_path = path
    image_patch_path = image_patch_path.replace(
        "masks/", dataset_type + "_masks/" + subdir + "/"
    )

    if i == 0:
        clear_destination_folder(image_patch_path)

    # print('mask save loacation: ' + image_patch_path)
    for j, patch in enumerate(patches):
        image_patch_path_numbered = f"{image_patch_path[:-4]}_{j}.tif"
        cv2.imwrite(image_patch_path_numbered, patch)


def process_image_and_mask(
    im_path: str,
    model_name: str,
    root_path: str,
    dataset_type: str,
    patch_size: int,
    patch_dir: str,
    i: int,
    scaling_factor: float,
    file_list: List[str],
    split: bool,
) -> None:
    """
    Used by patch_and_save, not ment to interact directly but possible to create
    own function based on this function.
    Process an image and its corresponding mask, create patches, and save them.

    Parameters:
        - im_path (str): Path to the original image file.
        - model_name (str): Name of the model or dataset.
        - dataset_type (str): Type of the dataset (e.g., 'train', 'test').
        - patch_size (int): Size of the patches to be created.
        - patch_dir (str): Directory where patches will be saved.
        - i (int): Index used for saving numbered patches.
        - scaling_factor (float): A factor to scale the images and masks.
        - file_list (list): List of file paths.

    Notes:
        - Loads image and masks, crops them to a petri dish area, and pads them to match patch size.
        - Rescales the images and masks if specified.
        - Creates patches for the image and saves them.
        - Reshapes the mask and passes it to save_mask_patches.

    Raises:
        ValueError: If either image fails to load.
    """

    # Extract image name by removing path and extension
    im_name = os.path.basename(im_path)
    im_name = im_name.replace(".png", "")

    # Get masks paths
    root_mask_path = root_path + model_name + "/masks/" + im_name + "_root_mask.tif"
    logger.debug(f"Processing image name: {im_name}")
    logger.debug(f"Processing root mask path: {root_mask_path}")
    logger.debug(f"Patch size: {patch_size}")

    # Load image and masks
    # Load grayscale for croping to petri dish
    im = cv2.imread(im_path, cv2.IMREAD_GRAYSCALE)
    root_mask = cv2.imread(root_mask_path, cv2.IMREAD_GRAYSCALE)

    # Check if the images were loaded successfully
    if im is None:
        raise ValueError(f"Failed to load image from path: {im_path}")
    if root_mask is None:
        raise ValueError(f"Failed to load root mask from path: {root_mask_path}")

    # Crop to petri dish using grayscale image
    _, roiX, roiY, roiW, roiH = crop_to_petri(im)
    # Crop image and masks to petri dish
    im = im[roiY : roiY + roiH, roiX : roiX + roiW]
    root_mask = root_mask[roiY : roiY + roiH, roiX : roiX + roiW]

    # Pad images and masks to be devidabe by patch size
    im = padder(im, patch_size)
    root_mask = padder(root_mask, patch_size)

    # Check if the shape of image and mask are corresponding
    if i == 0:
        if im.shape[0] == root_mask.shape[0] and im.shape[1] == root_mask.shape[1]:
            logger.info(f"Image size X: {im.shape[0]} Y: {im.shape[1]}")
            expected_files = (
                len(file_list) * (im.shape[0] / patch_size) * (im.shape[1] / patch_size)
            )

            logger.info(f"expected number of files: {expected_files}")
    elif i == 0:
        logger.error("ERROR: size don't match")
        return 1

    # Rescale if specified
    if scaling_factor != 1:
        logger.info(f"Rescaling with factor of {scaling_factor}")
        im = cv2.resize(im, (0, 0), fx=scaling_factor, fy=scaling_factor)
        root_mask = cv2.resize(root_mask, (0, 0), fx=scaling_factor, fy=scaling_factor)

    # Create patches for image
    patches = patchify(im, (patch_size, patch_size), step=patch_size)
    patches = patches.reshape(-1, patch_size, patch_size)

    # Save patches for image

    # TO-DO: Save in correct place
    # e.g. im_path: ../data/test/test/image.png
    # e.g. image_patch_path: ../data/test/test_images/test/image.png
    image_patch_path = im_path.replace(root_path, patch_dir)
    image_patch_path = image_patch_path.replace("\\", "/")

    # TO-DO: Fix path
    if split:
        image_patch_path = image_patch_path.replace(
            model_name + "/" + dataset_type + "/",
            model_name + "/" + dataset_type + "_images/" + dataset_type + "/",
        )
    else:
        image_patch_path = image_patch_path.replace(
            model_name + "/images/",
            model_name + "/" + dataset_type + "_images/" + dataset_type + "/",
        )

    logger.debug("Image patch path: " + image_patch_path)
    # print('=====')
    # print('original image path: ' + im_path)
    # print('model_name: ' + model_name)
    # print('dataset_type: ' + dataset_type)
    # print('image save location: ' + image_patch_path)
    if i == 0:
        clear_destination_folder(image_patch_path)

    for j, patch in enumerate(patches):
        image_patch_path_numbered = f"{image_patch_path[:-4]}_{j}.png"
        cv2.imwrite(image_patch_path_numbered, patch)

    # Reshape the mask
    root_mask = root_mask.reshape(root_mask.shape[0], root_mask.shape[1], 1)

    save_mask_patches(
        root_mask_path,
        root_mask,
        patch_size,
        root_path,
        "root",
        patch_dir,
        dataset_type,
        i,
    )
    return 0


def patch_and_save(
    model_name: str = "test",
    root_path: str = "./data/",
    patch_size: int = 256,
    scaling_factor: float = 1,
    patch_dir: str = "./data_patched/",
    test_split: float = 0.3,
    # uri: str = None,
    # azure: bool = False # OUTPUT
) -> None:
    """
    Generate and save patches for images and corresponding masks.

    Parameters:
        - model_name (str): Name of the model or dataset.
        - root_path (str): Path to the root of data folder.
        - patch_size (int): Size of the patches to be created.
        - scaling_factor (float): A factor to scale the images and masks.
        - patch_dir (str): Directory to save the patched images.
        - test_split (float): A split for test dataset if not split into train/test before upload.

    Returns:
        0 if the process is succesfull, 1 if an error occurs

    Note:
        This function assumes a specific dataset structure with separate subdirectories for images and masks.

    .. code-block::

        data/
        │
        ├── train/
        │   ├── Image1.png
        │   ├── Image2.png
        │   └── ...
        ├── test/
        │   ├── Image1.png
        │   ├── Image2.png
        │   └── ...
        └── masks/
            ├── Mask1_root_mask.tif
            ├── Mask1_shoot_mask.tif
            ├── ...
            ├── Mask2_root_mask.tif
            ├── Mask2_shoot_mask.tif
            └── ...

    The generated image and mask patches are saved in the specified patch_dir.

    """
    # ------------------------------------- #
    # TO-DO: Load data as separate function #
    # ------------------------------------- #
    # Get all image names in train_images
    folder_list = glob.glob(root_path + model_name + "/*")
    abs_folder_list = [os.path.abspath(folder) for folder in folder_list]
    # logger.info(f'Exploring {abs_folder_list} for images and masks folders')
    logger.info(f"Exploring {folder_list} for images and masks folders")

    folder_list = [file.replace("\\", "/") for file in folder_list]

    folder_list = [
        file.replace(root_path + model_name + "/", "") for file in folder_list
    ]

    logger.debug(folder_list)
    valid_names_train_test = ["train", "test", "masks"]
    valid_names_images_masks = ["masks", "images"]

    # Detect folder structure and separate files into train and test
    if all(name in folder_list for name in valid_names_train_test):
        # log("detected train, test, and masks.")
        logger.info("Processing pre-split dataset")

        train_folder = root_path + model_name + "/train"
        test_folder = root_path + model_name + "/test"
        masks_folder = root_path + model_name + "/masks"

        logger.info("Cleaining folders")
        compare_and_remove(train_folder, masks_folder)
        compare_and_remove(test_folder, masks_folder)
        logger.info("Folders cleaned")
        # Get a list of all files for train and test
        # Masks will be handled later alongside images
        train_file_list = glob.glob(train_folder + "/*")
        test_file_list = glob.glob(test_folder + "/*")
        split = True

    elif all(name in folder_list for name in valid_names_images_masks):
        # log("detected masks and images.")
        logger.info("Processing not pre-split dataset")
        images_folder = root_path + model_name + "/images"
        masks_folder = root_path + model_name + "/masks"

        logger.info("Cleaning folders")
        compare_and_remove(images_folder, masks_folder)
        logger.info("Folders cleaned")

        file_list = glob.glob(images_folder + "/*")

        # Get number of items to split into train/test
        num_files = len(file_list)
        logger.debug(f"Number of files in images folder: {num_files}")
        logger.debug(f"Test split: {test_split}")
        test_split_number = int(num_files * test_split)

        # Get test split
        test_file_list = random.sample(file_list, test_split_number)
        # Get train split, left over from test split
        train_file_list = [file for file in file_list if file not in test_file_list]
        split = False

    else:
        print("The file names do not match the expected pattern.")
        logger.info("The file names do not match the expected pattern.")

        return 1
    # ------------------------------------- #
    # TO-DO: Load data as separate function #
    # ------------------------------------- #
    logger.info("Processing train")
    i = 0
    for im_path in train_file_list:
        response = process_image_and_mask(
            im_path,
            model_name,
            root_path,
            "train",
            patch_size,
            patch_dir,
            i,
            scaling_factor,
            train_file_list,
            split,
        )
        if response == 1:
            return 1
        i += 1
    logger.info("Processing test")
    i = 0
    for im_path in test_file_list:
        response = process_image_and_mask(
            im_path,
            model_name,
            root_path,
            "test",
            patch_size,
            patch_dir,
            i,
            scaling_factor,
            train_file_list,
            split,
        )
        if response == 1:
            return 1
        i += 1
    # TO-DO: if azure and uri upload to uri
    return 0


class DataGenerator(Sequence):
    def __init__(self, image_gen, mask_gen):
        self.image_gen = image_gen
        self.mask_gen = mask_gen

    def __len__(self):
        return min(len(self.image_gen), len(self.mask_gen))

    def __getitem__(self, index):
        x = self.image_gen[index]
        y = self.mask_gen[index]
        return x[0], y[0]


def load_and_preprocess_data(
    classes: List[str],
    model_name: str = "test",
    patch_size: int = 256,
    patch_dir: str = "./data_patched/",
    seed: int = 42,
    batch_size: int = 16,
    image_color_mode: str = "grayscale",
) -> Tuple[DataGenerator, DataGenerator, int, int]:
    """
    Load and preprocess image and mask data for training and testing.

    Parameters:
        - classes (List[str]): List of class names for masks.
        - model_name (str): Name of the model or dataset. Default is "test".
        - patch_size (int): Size of the patches. Default is 256.
        - patch_dir (str): Directory containing image and mask patches. Default is "data_patched".
        - seed (int, optional): Seed for data generators. Default is 42.
        - batch_size (int): Batch size for the generators. Default is 16.
        - image_color_mode (str): Color mode for images. Default is 'grayscale'.

    Returns:
        Tuple[zip, zip, int, int]: A tuple containing:
            - train_generator: Training data generator (image and mask).
            - test_generator: Testing data generator (image and mask).
            - steps_per_epoch: Number of steps to be taken per epoch.
            - validation_steps: Number of steps to be taken for validation.

    Notes:
        - The function uses ImageDataGenerator to create data generators for image and mask patches.
        - Training and testing data generators are created for both images and masks.
    """

    logger.info("Initializing data generators for training and testing.")

    logger.debug("Creating ImageDataGenerator for training images.")
    train_image_datagen = ImageDataGenerator(rescale=1.0 / 255)

    train_image_generator = train_image_datagen.flow_from_directory(
        os.path.join(
            patch_dir,
            model_name,
            "train_images",
        ),
        target_size=(patch_size, patch_size),
        batch_size=batch_size,
        color_mode=image_color_mode,
        seed=seed,
    )

    logger.debug("Creating ImageDataGenerator for training masks.")
    train_mask_datagen = ImageDataGenerator()

    train_mask_generator = train_mask_datagen.flow_from_directory(
        os.path.join(patch_dir, model_name, "train_masks"),
        target_size=(patch_size, patch_size),
        batch_size=batch_size,
        classes=classes,
        color_mode="grayscale",
        seed=seed,
    )

    train_generator = DataGenerator(train_image_generator, train_mask_generator)

    logger.debug("Creating ImageDataGenerator for testing images.")
    test_image_datagen = ImageDataGenerator(rescale=1.0 / 255)

    test_image_generator = test_image_datagen.flow_from_directory(
        os.path.join(patch_dir, model_name, "test_images"),
        target_size=(patch_size, patch_size),
        batch_size=batch_size,
        color_mode=image_color_mode,
        seed=seed,
    )

    logger.debug("Creating ImageDataGenerator for testing masks.")
    test_mask_datagen = ImageDataGenerator()

    test_mask_generator = test_mask_datagen.flow_from_directory(
        os.path.join(patch_dir, model_name, "test_masks"),
        target_size=(patch_size, patch_size),
        batch_size=batch_size,
        classes=classes,
        color_mode="grayscale",
        seed=seed,
    )

    test_generator = DataGenerator(test_image_generator, test_mask_generator)

    steps_per_epoch = len(train_image_generator)
    validation_steps = test_image_generator.samples // batch_size

    logger.info(
        f"Data generators created successfully. Steps per epoch: {steps_per_epoch}, Validation steps: {validation_steps}"
    )

    return train_generator, test_generator, steps_per_epoch, validation_steps


def load_and_preprocess_image(
    im_path: str,
    patch_size: int = 256,
    scaling_factor: float = 1,
    num_channels: int = 1,
) -> Tuple[np.ndarray, int, int, np.ndarray]:
    """
    Load, crop, preprocess, and create patches from an image.

    Parameters:
        - im_path (str): Path to the image.
        - patch_size (int): Size of the patches to create.
        - scaling_factor (float): Scaling factor for the image.
        - num_channels (int): Number of channels for the image (1 for grayscale, 3 for color).

    Returns:
        - patches (np.ndarray): Patches created from the image.
        - i (int): Number of patches in the vertical dimension.
        - j (int): Number of patches in the horizontal dimension.
        - im (np.ndarray): Preprocessed image.
    """

    logger.info(f"Loading image from path: {im_path}")
    if num_channels == 1:
        im = cv2.imread(im_path, cv2.IMREAD_GRAYSCALE)
    else:
        im = cv2.imread(im_path, cv2.IMREAD_COLOR)

    if im is None:
        logger.error("Error loading image.")
        raise ValueError("Error loading image")

    if num_channels != 1:
        im_gray = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
    else:
        im_gray = im

    logger.info("Cropping image to petri dish.")
    _, roiX, roiY, roiW, roiH = crop_to_petri(im_gray)
    cropped_im = im[roiX : roiX + roiW, roiY : roiY + roiH]

    logger.info("Padding cropped image.")
    padded_im = padder(cropped_im, patch_size)

    if scaling_factor != 1:
        logger.info(f"Scaling image by a factor of {scaling_factor}.")
        padded_im = cv2.resize(padded_im, (0, 0), fx=scaling_factor, fy=scaling_factor)

    logger.info("Creating patches from the padded image.")
    patches = patchify(padded_im, (patch_size, patch_size))
    i = patches.shape[0]
    j = patches.shape[1]
    patches = patches.reshape(-1, patch_size, patch_size, 1 if num_channels == 1 else 3)

    logger.info("Image preprocessing completed.")
    return patches, i, j, padded_im


def main():
    parser = argparse.ArgumentParser(
        description="Preprocess data using predefinded folder struncture, /data/model_name/ train, test, masks or images, masks. If more flexible solution is needed please use process_image_and_mask function in this module to create a custom loop"
    )
    parser.add_argument(
        "model_name",
        default="test",
        help="Name of the model that data is being preprocessed.",
        type=str,
    )
    parser.add_argument(
        "-r",
        "--root_path",
        default="./data/",
        action="store",
        help="Path to the root of data folder. End with /.",
        type=str,
    )
    parser.add_argument(
        "-p",
        "--patch_size",
        default=256,
        action="store",
        help="What patches to use douring preprocessing. Default is 256.",
        type=int,
    )
    parser.add_argument(
        "-s",
        "--scaling_factor",
        default=1.0,
        action="store",
        help="Scaling factor of the images on the train and test dataset. Recommended 1",
        type=float,
    )
    parser.add_argument(
        "-d",
        "--patch_dir",
        default="./data_patched/",
        action="store",
        help="Root path for patched data for training. WARNING Changes to this variable may result in broken pipe. End with /.",
        type=str,
    )
    parser.add_argument(
        "-t",
        "--test_split",
        default=0.3,
        action="store",
        help="Test split if the dataset is not alerady split into train and test dataset",
        type=float,
    )

    # TO-DO: Add azure and uri paths

    args = parser.parse_args()

    patch_and_save(
        args.model_name,
        args.root_path,
        args.patch_size,
        args.scaling_factor,
        args.patch_dir,
        args.test_split,
    )


if __name__ == "__main__":
    main()
